
/*Utility routines for Jdat and Fdat types*/

#include "local.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "son.h"
#include "../em/em.typ"
#include "jxy.typ"
#include <ctype.h>

/* char *strcpy(),*strncpy(); KRS */

extern char prog_name[];

clear_pdat(pd) Pdat *pd;
/*************************************************************/
/* Wipe out all mem allocated for pd list, incl. pd.         */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr19apr91 */ 
{ Pdat *pd_next;

  clear_pdat_mem(pd);  /*clears all data recursively*/

  while (pd)
  { pd_next = pd->next;
    sFree((char *)pd);
    pd = pd_next;
  }

} /*clear_pdat*/

clear_pdat_mem(pd) Pdat *pd;
/*************************************************************/
/* Free all allocated memeory in pd and rest of pd list.     */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr19apr91 */ 
{ 
  while (pd)
  { if (pd->jx_mg!=NULL) sFree((char *)pd->jx_mg);  pd->jx_mg = NULL;
    if (pd->jx_ph!=NULL) sFree((char *)pd->jx_ph);  pd->jx_ph = NULL;
    if (pd->jy_mg!=NULL) sFree((char *)pd->jy_mg);  pd->jy_mg = NULL;
    if (pd->jy_ph!=NULL) sFree((char *)pd->jy_ph);  pd->jy_ph = NULL;
    if (pd->ez_mg!=NULL) sFree((char *)pd->ez_mg);  pd->ez_mg = NULL;
    if (pd->ez_ph!=NULL) sFree((char *)pd->ez_ph);  pd->ez_ph = NULL;
    if (pd->stats!=NULL) sFree((char *)pd->stats);  pd->stats = NULL;
    pd = pd->next;
  }

} /*clear_pdat_mem*/

clear_tdat(td) Tdat *td;
/*************************************************************/
/* Free all allocated memeory in rd.                         */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01jun93 */ 
{ 
  td->pn = 0;
  td->snapped = 0;
  if (td->tri!=NULL)
  { clear_ipts(td->tri);       td->tri = NULL; }
  if (td->jx_mg!=NULL)
  { sFree((char *)td->jx_mg);  td->jx_mg = NULL; }
  if (td->jx_ph!=NULL)
  { sFree((char *)td->jx_ph);  td->jx_ph = NULL; }
  if (td->jy_mg!=NULL)
  { sFree((char *)td->jy_mg);  td->jy_mg = NULL; }
  if (td->jy_ph!=NULL)
  { sFree((char *)td->jy_ph);  td->jy_ph = NULL; }
  if (td->ez_mg!=NULL)
  { sFree((char *)td->ez_mg);  td->ez_mg = NULL; }
  if (td->ez_ph!=NULL)
  { sFree((char *)td->ez_ph);  td->ez_ph = NULL; }

} /*clear_tdat*/

clear_jdat(jd) Jdat *jd;
/*************************************************************/
/* Close down all allocated memory in jd.                    */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ int i;

  if (jd==NULL)
    announce_bug("clear_jdat");

  /*Shut down all Fdat structures*/
  if (jd->fdat)
  { clear_fdat(jd->fdat);  /*recursively clears all fdat's*/
    sFree((char *)jd->fdat);  /*Allocated as an array, this frees all of them at once*/
  }
  if (jd->freq) sFree((char *)jd->freq);  jd->freq = NULL;

  /*Shut down rectangle storage arrays (one per level)*/
  if (jd->rect)
  { for (i=0; i<jd->cir->n-1; i++)
      clear_ipts( jd->rect+i );
    sFree((char *)jd->rect);  jd->rect = NULL;
  }
  if (jd->xterp) sFree((char *)jd->xterp);  jd->xterp = NULL;
  if (jd->yterp) sFree((char *)jd->yterp);  jd->yterp = NULL;

  if (jd->vsrc_mg) sFree((char *)jd->vsrc_mg);  jd->vsrc_mg = NULL;
  if (jd->vsrc_ph) sFree((char *)jd->vsrc_ph);  jd->vsrc_ph = NULL;
  if (jd->rsrc) sFree((char *)jd->rsrc);  jd->rsrc = NULL;
  if (jd->xsrc) sFree((char *)jd->xsrc);  jd->xsrc = NULL;
  if (jd->lsrc) sFree((char *)jd->lsrc);  jd->lsrc = NULL;
  if (jd->csrc) sFree((char *)jd->csrc);  jd->csrc = NULL;

  if (jd->cmd) sFree((char *)jd->cmd);  jd->cmd = NULL;
  if (jd->nam) sFree((char *)jd->nam);  jd->nam = NULL;
  if (jd->tim) sFree((char *)jd->tim);  jd->tim = NULL;
  if (jd->ver) sFree((char *)jd->ver);  jd->ver = NULL;

  if (jd->box)
  { clean_box(jd->box);
    sFree((char *)jd->box);  jd->box = NULL;
  }

  if (jd->cir)
  { clear_cir(jd->cir);
    sFree((char *)jd->cir);  jd->cir = NULL;
  }

  jd->nport = jd->nsub = jd->nfreq = 0;

  /*Recursively clear any additional jdats, de-allocate memory*/
  if (jd->next)
  { clear_jdat(jd->next);
    sFree((char *)jd->next);
    jd->next = NULL;
  }

} /*clear_jdat*/

clear_fdat(fd) Fdat *fd;
/*************************************************************/
/* Close down all allocated memory in fd.                    */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ 
  if (fd==NULL)
    announce_bug("clear_fdat");

  if (fd->jpr) Vec_close(fd->jpr);  fd->jpr = NULL;
  if (fd->ypr) Mat_close(fd->ypr);  fd->ypr = NULL;
  if (fd->cans) Mat_close(fd->cans);  fd->cans = NULL;
  if (fd->Z0) sFree((char *)fd->Z0);  fd->Z0 = NULL;
  if (fd->Eeff) sFree((char *)fd->Eeff);  fd->Eeff = NULL;

  fd->nsub = fd->nport = 0;
  fd->freq = 0.0;

  /*Recursively clear all remaining fdats.*/
  if (fd->next) clear_fdat(fd->next);

} /*clear_fdat*/

init_pdat(pd) Pdat *pd;
/*************************************************************/
/* Set values to reasonable start up numbers.                */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr19apr91 */ 
{ pd->x0 = pd->y0 = 0.0;
  pd->nx = pd->ny = 0;
  pd->dx = pd->dy = 0.0;
  pd->ira = pd->irb = NULL;
  pd->pvar = STEP_F;
  pd->pport = 0;
  pd->plist = NULL;
  pd->pn = 0;
  pd->jx_mg = pd->jx_ph = NULL;
  pd->jy_mg = pd->jy_ph = NULL;
  pd->ez_mg = pd->ez_ph = NULL;
  pd->stats = NULL;
  pd->next = NULL;
} /*init_pdat*/

init_pdat_stats(pd) Pdat *pd;
/*************************************************************/
/* Set stats values to reasonable start up numbers.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr17may91 */ 
{ int i;
  Rtyp *stats;

  if (pd->stats!=NULL) /*stats must be free'd each time*/
    announce_bug("init_pdat_stats");

  pd->stats = (Rtyp *)new( pd->pn * NUM_STATS * sizeof(Rtyp) );

  for (i=0,stats=pd->stats; i<pd->pn; i++,stats+=NUM_STATS)
  { *(stats + JX_MG_MAX) = 0.0;
    *(stats + JX_MG_MIN) = BIG;
    *(stats + JX_MG_AVE) = 0.0;
    *(stats + JX_MG_SIG) = 0.0;
    *(stats + JX_PH_MAX) = -PI;
    *(stats + JX_PH_MIN) = PI;
    *(stats + JX_PH_AVE) = 0.0;
    *(stats + JX_PH_SIG) = 0.0;

    *(stats + JY_MG_MAX) = 0.0;
    *(stats + JY_MG_MIN) = BIG;
    *(stats + JY_MG_AVE) = 0.0;
    *(stats + JY_MG_SIG) = 0.0;
    *(stats + JY_PH_MAX) = -PI;
    *(stats + JY_PH_MIN) = PI;
    *(stats + JY_PH_AVE) = 0.0;
    *(stats + JY_PH_SIG) = 0.0;

    *(stats + EZ_MG_MAX) = 0.0;
    *(stats + EZ_MG_MIN) = BIG;
    *(stats + EZ_MG_AVE) = 0.0;
    *(stats + EZ_MG_SIG) = 0.0;
    *(stats + EZ_PH_MAX) = -PI;
    *(stats + EZ_PH_MIN) = PI;
    *(stats + EZ_PH_AVE) = 0.0;
    *(stats + EZ_PH_SIG) = 0.0;
  }
} /*init_pdat_stats*/

init_tdat(td) Tdat *td;
/*************************************************************/
/* Set values to reasonable start up numbers.                */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01jun93 */ 
{ td->tri = NULL;
  td->pn = 0;
  td->snapped = 0;
  td->jx_mg = td->jx_ph = NULL;
  td->jy_mg = td->jy_ph = NULL;
  td->ez_mg = td->ez_ph = NULL;
} /*init_tdat*/

init_jdat(jd)  Jdat *jd;
/*************************************************************/
/* Initialize jd, prepare for file read.                     */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ int i;

  if ( jd==NULL )
    announce_bug("init_jdat");

  jd->nport = 0;
  jd->nsub = 0;
  jd->nfreq = 0;

  jd->box = (Box_info *)new( sizeof(Box_info) );
  init_box(jd->box);

  jd->cir = (Cir *)cnew( 1,sizeof(Cir) );
  /*cir is initialized when file read by Get_circuit(box,cir,1)*/

  /*Character strings from running em, read from .jxy file*/
  jd->cmd = jd->nam = jd->tim = jd->ver = NULL;

  /*One cmplx voltage src and src impedance for each port*/
  jd->vsrc_mg = NULL;
  jd->vsrc_ph = NULL;
  jd->rsrc = NULL;
  jd->xsrc = NULL;
  jd->lsrc = NULL;
  jd->csrc = NULL;

  /*One rectangle list will be allocated for each level*/
  jd->rect = NULL;
  /*Interpolation coordinates for msub->loc*/
  jd->xterp = jd->yterp = NULL;

  /*Frequency list will be stored here*/
  jd->freq = NULL;

  /*One fdat for each frequency*/
  /*Will set up as a linked list, then counted to set nfreq,
    then freq and jd->fdat allocated as arrays.  jd->fdat->next
    ptr will be kept intact when fdat is moved into jd->fdat
    array.
  */
  jd->fdat = NULL;

  /*Pointer to next jdat in list*/
  jd->next = NULL;

} /*init_jdat*/

init_fdat(fd,nport,nsub)  Fdat *fd;  int nport,nsub;
/*************************************************************/
/* Initialize and allocate all required memory in fd.        */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ 
  if ( fd==NULL || nport<0 || nsub<0 )
    announce_bug("init_fdat");

  fd->nport = nport;
  fd->nsub = nsub;
  fd->freq = 0.0;

  if (nport>0)
  { fd->Eeff = (Rtyp *)cnew( 2*nport,sizeof(Rtyp) );
    fd->Z0   = (Rtyp *)cnew( 2*nport,sizeof(Rtyp) );
    fd->cans = Mat_open(nport,nport);
  }
  if (nport>0 && nsub>0) fd->ypr  = Mat_open(nport,nsub);
  if (nsub>0)  fd->jpr  = Vec_open(nsub);

  fd->next = NULL;

} /*init_fdat*/

copy_fdat(fd1,fd2) Fdat *fd1,*fd2;
/*************************************************************/
/* Copy from 2 into 1.                                       */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ 
  fd1->nport = fd2->nport;
  fd1->nsub = fd2->nsub;
  fd1->freq = fd2->freq;
  fd1->Eeff = fd2->Eeff;
  fd1->Z0 = fd2->Z0;
  fd1->cans = fd2->cans;
  fd1->ypr = fd2->ypr;
  fd1->jpr = fd2->jpr;
} /*copy_fdat*/

cmp_jrect(xya,xyb)  int xya[4],xyb[4];
/*************************************************************/
/* Compare two rectangles for sorting.                       */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr23apr91 */ 
{
  if (xya[1] < xyb[1]) return(-1);
  if (xya[1] > xyb[1]) return(1);

  if (xya[0] < xyb[0]) return(-1);
  if (xya[0] > xyb[0]) return(1);

  if (xya[3] < xyb[3]) return(-1);
  if (xya[3] > xyb[3]) return(1);

  if (xya[2] < xyb[2]) return(-1);
  if (xya[2] > xyb[2]) return(1);

  return(0);  /*They are equal*/
} /*cmp_jrect*/

setup_jdat(jd)  Jdat *jd;
/*************************************************************/
/* With file read complete, set up rect, etc.                */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr17apr91 */ 
{ int i,j,ilev,nlev,loc;
  Ipts *rect;
  msub_list *lev;
  msub *im,*jm,*mstop;
  int ikey[RSIZE];
  int *ir,*jr,*ir_stop;
  int ck_rect(),cmp_all_ints();

  /*Exclude ground for number of levels*/
  nlev = jd->cir->n - 1;

  /*Store x,y interpolation coordinates for each location*/
  for (ilev=0; ilev<nlev; ilev++)
  { lev = jd->cir->lev0 + ilev;
    mstop = lev->msub0 + lev->n;
    for (im=lev->msub0; im<mstop; im++) switch(im->id)
    { case MXR:
      case MYR:   *(jd->xterp+im->loc) = *(im->ip+MX2);
                  *(jd->yterp+im->loc) = *(im->ip+MY2);
                  break;
      case MTNN:  *(jd->xterp+im->loc) = *(im->ip+MX0) + 1;
                  *(jd->yterp+im->loc) = *(im->ip+MY0) - 1;
                  break;
      case MTNP:  *(jd->xterp+im->loc) = *(im->ip+MX0) + 1;
                  *(jd->yterp+im->loc) = *(im->ip+MY0) + 1;
                  break;
      case MTPN:  *(jd->xterp+im->loc) = *(im->ip+MX0) - 1;
                  *(jd->yterp+im->loc) = *(im->ip+MY0) - 1;
                  break;
      case MTPP:  *(jd->xterp+im->loc) = *(im->ip+MX0) - 1;
                  *(jd->yterp+im->loc) = *(im->ip+MY0) + 1;
                  break;
      case MENDYN:
      case MENDYP:
      case MENDXN:
      case MENDXP: *(jd->xterp+im->loc) = *(im->ip+MX0);
                   *(jd->yterp+im->loc) = *(im->ip+MY0);
                   break;
    } /*for im, switch*/
  } /*for lev*/

  /*One rect list for each level*/
  jd->rect = (Ipts *)new( nlev*sizeof(Ipts) ); 
  for (i=0; i<nlev; i++)
  { rect = jd->rect+i;
    init_ipts(rect,RSIZE);
    /*Set comparison fucntion and rules for building list*/
    rect->cmp_func = cmp_jrect;
    rect->rules = ck_rect;
  }

  /*Build sorted, duped list of rectangles:
    ikey[0] = x0, ikey[1] = y0, ikey[2] = x1, ikey[3] = y1,
    ikey[4] = index of lower overlapping msub, Jx,
    ikey[5] = index of higher overlapping msub, Jx.
    ikey[6] = index of lower overlapping msub, Jy,
    ikey[7] = index of higher overlapping msub, Jy.
    [4] - [7] set after rect list is built.
    -1 indicates no overlapping msub.
  */
  ikey[4] = ikey[5] = ikey[6] = ikey[7] = -1;
  for (ilev=0; ilev<nlev; ilev++)
  { lev = jd->cir->lev0 + ilev;
    rect = jd->rect + ilev;
    mstop = lev->msub0 + lev->n;
    for (im=lev->msub0; im<mstop; im++)
    { if (im->id==MXR)
      { ikey[0] = *im->ip;  ikey[1] = *(im->ip+1);
        ikey[2] = *(im->ip+4);  ikey[3] = *(im->ip+3);
        if (ikey[0]>=0) add_ipts(rect,ikey);
        ikey[0] = ikey[2];  ikey[2] = *(im->ip+2);
        if (ikey[2]<=jd->box->Capm) add_ipts(rect,ikey);
      }
      else if (im->id==MYR)
      { ikey[0] = *im->ip;  ikey[1] = *(im->ip+1);
        ikey[2] = *(im->ip+2);  ikey[3] = *(im->ip+5);
        if (ikey[1]>=0) add_ipts(rect,ikey);
        ikey[1] = ikey[3];  ikey[3] = *(im->ip+3);
        if (ikey[3]<=jd->box->Capn) add_ipts(rect,ikey);
      }
      else if (im->id==MTPP)
      { ikey[0] = *(im->ip)-3;  ikey[1] = *(im->ip+1)-1;
        ikey[2] = *(im->ip)-1;  ikey[3] = *(im->ip+1)+1;
        if (ikey[0]>=0) add_ipts(rect,ikey);
        ikey[0] = ikey[2];  ikey[1] = ikey[3];
        ikey[2] = *(im->ip)+1;  ikey[3] = *(im->ip+1)+3;
        if (ikey[3]<=jd->box->Capn) add_ipts(rect,ikey);
      }
      else if (im->id==MTPN)
      { ikey[0] = *(im->ip)-3;  ikey[1] = *(im->ip+1)-1;
        ikey[2] = *(im->ip)-1;  ikey[3] = *(im->ip+1)+1;
        if (ikey[0]>=0) add_ipts(rect,ikey);
        ikey[0] = *(im->ip)-1;  ikey[1] = *(im->ip+1)-3;
        ikey[2] = *(im->ip)+1;  ikey[3] = *(im->ip+1)-1;
        if (ikey[1]>=0) add_ipts(rect,ikey);
      }
      else if (im->id==MTNP)
      { ikey[0] = *(im->ip)+1;  ikey[1] = *(im->ip+1)-1;
        ikey[2] = *(im->ip)+3;  ikey[3] = *(im->ip+1)+1;
        if (ikey[2]<=jd->box->Capm) add_ipts(rect,ikey);
        ikey[0] = *(im->ip)-1;  ikey[1] = *(im->ip+1)+1;
        ikey[2] = *(im->ip)+1;  ikey[3] = *(im->ip+1)+3;
        if (ikey[3]<=jd->box->Capn) add_ipts(rect,ikey);
      }
      else if (im->id==MTNN)
      { ikey[0] = *(im->ip)+1;  ikey[1] = *(im->ip+1)-1;
        ikey[2] = *(im->ip)+3;  ikey[3] = *(im->ip+1)+1;
        if (ikey[2]<=jd->box->Capm) add_ipts(rect,ikey);
        ikey[0] = *(im->ip)-1;  ikey[1] = *(im->ip+1)-3;
        ikey[2] = *(im->ip)+1;  ikey[3] = *(im->ip+1)-1;
        if (ikey[1]>=0) add_ipts(rect,ikey);
      }
    } /*for im*/

    /*resize/add all partially overlapping rectangles*/
    for (i=0; i<rect->n; i++)
    { for (j=i+1; j<rect->n; j++)
      { ir = rect->ipt0 + i * rect->size;  /*ipt0 might chng inside loop*/
        jr = rect->ipt0 + j * rect->size;
        /*Check for ir only partially contained by jr in y dir*/
        if ( *ir == *jr && *(ir+2) == *(jr+2) )
        { if ( *(ir+1) < *(jr+1) && *(ir+3) > *(jr+1) && *(ir+3) < *(jr+3) )
          { ikey[0] = *ir;  ikey[2] = *(ir+2);  /*Setting up middle rectangle*/
            ikey[1] = *(jr+1);  ikey[3] = *(ir+3);
            *(ir+3) = ikey[1];  /*Lower y rectangle*/
            *(jr+1) = ikey[3];  /*Upper y rectangle*/
            add_ipts(rect,ikey);
            j = 1;
          }
          else if (*(ir+3) > *(jr+3) && *(ir+1) > *(jr+1) && *(ir+1) < *(jr+3) )
          { ikey[0] = *jr;  ikey[2] = *(jr+2);  /*Setting up middle rectangle*/
            ikey[1] = *(ir+1);  ikey[3] = *(jr+3);
            *(jr+3) = ikey[1];  /*Lower y rectangle*/
            *(ir+1) = ikey[3];  /*Upper y rectangle*/
            add_ipts(rect,ikey);
            j = 1;
          }
        }
        /*Check for ir only partially contained by jr in x dir*/
        else if ( *(ir+1) == *(jr+1) && *(ir+3) == *(jr+3) )
        { if ( *ir < *jr && *(ir+2) > *jr && *(ir+2) < *(jr+2) )
          { ikey[0] = *jr;  ikey[2] = *(ir+2);  /*Setting up middle rectangle*/
            ikey[1] = *(ir+1);  ikey[3] = *(ir+3);
            *(ir+2) = ikey[0];  /*Lower x rectangle*/
            *jr = ikey[2];  /*Upper x rectangle*/
            add_ipts(rect,ikey);
            j = 1;
          }
          else if (*(ir+2) > *(jr+2) && *ir > *jr && *ir < *(jr+2) )
          { ikey[0] = *ir;  ikey[2] = *(jr+2);  /*Setting up middle rectangle*/
            ikey[1] = *(ir+1);  ikey[3] = *(ir+3);
            *(jr+2) = ikey[0];  /*Lower x rectangle*/
            *ir = ikey[2];  /*Upper x rectangle*/
            add_ipts(rect,ikey);
            j = 1;
          }
        }
      } /*for j*/
    } /*for i*/

    /*Remove all zero area rects, shouldn't happen, this is
      included for robustness*/
    for (i=0; i<rect->n; i++)
    { ir = rect->ipt0 + i * rect->size;
      if ( *ir == *(ir+2) || *(ir+1) == *(ir+3) )
      { rm_ipts(rect,ir);
         i--;
      }
    }

    /*Remove all big rects which are covered by small rects*/
    for (i=0; i<rect->n; i++)
    { ir = rect->ipt0 + i * rect->size;
      for (j=i+1; j<rect->n; j++)
      { jr = rect->ipt0 + j * rect->size;
        /*Check for ir completely contained by jr, if so rm jr*/
        if ( *ir >= *jr && *(ir+2) <= *(jr+2) &&
             *(ir+1) >= *(jr+1) && *(ir+3) <= *(jr+3) )
        {   rm_ipts(rect,jr);
            j--;
        }
        /*Check for jr completely contained by ir, if so rm ir*/
        else if ( *jr >= *ir && *(jr+2) <= *(ir+2) &&
                  *(jr+1) >= *(ir+1) && *(jr+3) <= *(ir+3) )
        { rm_ipts(rect,ir);
          i--;
          break;
        }
      } /*for j*/
    } /*for i*/
  } /*for lev*/

  /*Go thru each rect on each level and find associated msubs*/
  for (i=0; i<nlev; i++)
  { lev = jd->cir->lev0 + i;
    rect = jd->rect + i;
    ir_stop = rect->ipt0 + rect->n * rect->size;
    for (ir=rect->ipt0; ir<ir_stop; ir+=rect->size)
    { /*Go thru each msub on level looking for adjacent msubs*/
      mstop = lev->msub0 + lev->n;
      for (im=lev->msub0; im<mstop; im++)
      { jr = im->ip;
        loc = im->loc;
        switch(im->id)
        { case MXR :
            if ( *(jr+MY0) > *(ir+RY0) || *(jr+MY1) < *(ir+RY1) )
	      break;
            if ( *(jr+MX2) <= *(ir+RX0) && *(jr+MX1) >= *(ir+RX1) )
	      *(ir+RXA) = loc;
            else if ( *(jr+MX2) >= *(ir+RX1) && *(jr+MX0) <= *(ir+RX0) )
	      *(ir+RXB) = loc;
            break;
          case MYR :
            if (*(jr+MX0) > *(ir+RX0) || *(jr+MX1) < *(ir+RX1) )
	      break;
            if ( *(jr+MY2) <= *(ir+MY0) && *(jr+MY1) >= *(ir+RY1) )
	      *(ir+RYA) = loc;
            else if ( *(jr+MY2) >= *(ir+MY1) && *(jr+MY0) <= *(ir+RY0) )
	      *(ir+RYB) = loc;
            break;
          case MTNN :
            if ( *jr == *(ir+RX0)-1 && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXA) = loc;
            else if ( *(jr+1) == *(ir+RY1)+1 && *jr == *(ir+RX0)+1 )
	      *(ir+RYB) = loc;
            break;
          case MTNP :
            if ( *jr == *(ir+RX0)-1 && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXA) = loc;
            else if ( *(jr+1) == *(ir+RY0)-1 && *jr == *(ir+RX0)+1 )
	      *(ir+RYA) = loc;
            break;
          case MTPN :
            if ( *jr == *(ir+RX1)+1 && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXB) = loc;
            else if ( *(jr+1) == *(ir+RY1)+1 && *jr == *(ir+RX0)+1 )
	      *(ir+RYB) = loc;
            break;
          case MTPP :
            if ( *jr == *(ir+RX1)+1 && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXB) = loc;
            else if ( *(jr+1) == *(ir+RY0)-1 && *jr == *(ir+RX0)+1 )
	      *(ir+RYA) = loc;
            break;
          case MENDYN :
            if ( *(jr+1) == *(ir+RY0) && *jr == *(ir+RX0)+1 )
	      *(ir+RYA) = loc;
            break;
          case MENDYP :
            if ( *(jr+1) == *(ir+RY1) && *jr == *(ir+RX0)+1 )
	      *(ir+RYB) = loc;
            break;
          case MENDXN :
            if ( *jr == *(ir+RX0) && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXA) = loc;
            break;
          case MENDXP :
            if ( *jr == *(ir+RX1) && *(jr+1) == *(ir+RY0)+1 )
	      *(ir+RXB) = loc;
            break;
        } /*swich im->id*/
      } /*for im*/
    } /*for ir*/
  } /*for i*/
} /*setup_jdat*/

update_jpr(jd,ifr)  Jdat *jd;  int ifr;
/*************************************************************/
/* Udate src info at ifr-th frequency.                       */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr17apr91 */ 
{ static Mat_ptr zy = NULL;
  static Vec_ptr tmp = NULL;
  static Vec_ptr vsrc = NULL;
  Rtyp zre,zim;
  Fdat *fd;
  int i;

  if (jd->vsrc_mg == NULL || jd->nport == 0) /*Must read file first!!!*/
    announce_bug("update_src");

  /*Update vsrc*/
  if ( vsrc!=NULL && jd->nport!=vsrc->Length )
  { Vec_close(vsrc);  vsrc = NULL; }
  if ( vsrc==NULL ) vsrc = Vec_open(jd->nport);

  for (i=0; i<jd->nport; i++)
  { zre = *(jd->vsrc_mg+i) * cos( *(jd->vsrc_ph+i) );
    zim = *(jd->vsrc_mg+i) * sin( *(jd->vsrc_ph+i) );
    Vec_put(i,zre,zim,vsrc);
  }

  if (zy != NULL && zy->Num_row != jd->nport)
  { Mat_close(zy);  zy = NULL; }
  if (zy==NULL) zy = Mat_open(jd->nport,jd->nport);

  fd = jd->fdat + ifr;
  for (i=0; i<jd->nport; i++)
  { calc_z(*(jd->rsrc+i),*(jd->xsrc+i),*(jd->lsrc+i),*(jd->csrc+i),fd->freq,&zre,&zim);
    Vec_smult(zre,zim,fd->cans->Row[i],zy->Row[i]);
  }

  for (i=0; i<jd->nport; i++)
    Mat_put_re(i,i,1.0+Mat_get_re(i,i,zy),zy);

  Mat_inv(zy,0);

  /*Make sure tmp vec is rite size*/
  if (tmp==NULL) tmp = Vec_open(jd->nsub);
  if (tmp->Length!=jd->nsub)
  { Vec_close(tmp);  tmp = Vec_open(jd->nsub); }

  for (i=0; i<jd->nport; i++)
  { Vec_dot(zy->Row[i],vsrc,&zre,&zim);
    if (i==0) Vec_smult(zre,zim,fd->ypr->Row[0],fd->jpr);
    else
    { Vec_smult(zre,zim,fd->ypr->Row[i],tmp);
      Vec_add(tmp,fd->jpr,fd->jpr);
    }
  }

} /*update_jpr*/

Pdat *setup_pdat(jd,ilev,xy,dx,dy)  Jdat *jd;  int ilev;  Rtyp xy[4],dx,dy;
/*************************************************************/
/* Return linked list ready for data to be calc'ed.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr18apr91 */
{ Pdat *pd0,*pd;
  Ipts *rect;
  int *ira,*irb,*ir_stop;
  int i,itt;
  Rtyp rtt,max;

  /*ilev = level number,
    xy[0] = x0, xy[1] = y0, xy[2] = x1, xy[3] = y1.
    if x0>x1 or y0>y1, order will be swapped.
    coordinates are 2 unit per cell.
    xy is region for returning data in pd0 (linked list).
    dx and dy are deltas for data calculation,
    usually set so one data point each pixel, dx = 0.1
    means 10 points per 1/2 cell (a cell is two integer coordinates).
  */

  /*Can't specify ground level (cir-n-1) for plotting*/
  if (ilev<0 || ilev>=jd->cir->n-1)
    announce_bug("setup_pdat(0)");

  if (xy[0]==xy[2] || xy[1]==xy[3]) return(NULL);

  /*Ck to make sure region pts are in rite order*/
  if (xy[0]>xy[2]) { rtt = xy[0];  xy[0] = xy[1];  xy[1] = rtt; }
  if (xy[1]>xy[3]) { rtt = xy[1];  xy[1] = xy[3];  xy[3] = rtt; }

  /*Search thru each rectangle looking for rects within desired region*/
  rect = jd->rect + ilev;
  ir_stop = rect->ipt0 + rect->n * rect->size;
  pd0 = NULL;
  for (ira=rect->ipt0; ira<ir_stop; ira+=rect->size) if (inside(ira,xy))
  { if (pd0==NULL)  pd = pd0 = (Pdat *)new( sizeof(Pdat) );
    else { pd->next = (Pdat *)new( sizeof(Pdat) );  pd = pd->next; }
    init_pdat(pd);

    /*See how far we can go with same rect y-dimension*/
    irb = ira + rect->size;
    while ( irb < ir_stop &&
            *(irb+RX1-RSIZE) == *(irb+RX0) &&
            *(ira+RY0) == *(irb+RY0) &&
            *(ira+RY1) == *(irb+RY1) &&
            inside(irb,xy) ) irb += rect->size;
    irb -= rect->size;
    /*ira is now the 1st valid rect for this region and irb is the last*/

    /***Initialize pd parameters and memory***/

    /*Calc zoom frame coord of corner of rect in cell units snapped to*/
    /*nearest inside pixel*/
    itt = (*(ira+RX0) - xy[0] + 0.99999*dx) / dx;
    pd->x0 = itt * dx;
    if (pd->x0 < 0.0) pd->x0 = 0.0;

    itt = (*(ira+RY0) - xy[1] + 0.99999*dy) / dy;
    pd->y0 = itt * dy;
    if (pd->y0 < 0.0) pd ->y0 = 0.0;

    pd->dx = dx;  pd->dy = dy;
    /*corner of zoom frame in substrate cell coord.*/
    pd->xx0 = xy[0];  pd->yy0 = xy[1];

    max = *(irb+RX1);
    if (max > xy[2]) max = xy[2];
    pd->nx = ( max - pd->x0 - xy[0] + 1.00001*dx) / dx;
    if (pd->nx <= 0) /*For robustness, should not be needed*/
      pd->nx = 1;

    max = *(irb+RY1);
    if (max > xy[3]) max = xy[3];
    pd->ny = ( max - pd->y0 - xy[1] + 1.00001*dy) / dy;
    if (pd->ny <= 0)  /*For robustness, should not be needed*/
      pd->ny = 1;

    pd->ira = ira;
    pd->irb = irb;
    ira = irb;

  } /*for ira*/

  return(pd0);
} /*setup_pdat*/

inside(xya,xyb)  int xya[4];  Rtyp xyb[4];
/*************************************************************/
/* Determine if any part of rect xya is inside xyb.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr22apr91 */ 
{
  /*xya/b stored as x0,y0,x1,y1, with x0<x1 && y0<y1*/

  if (xya[2]<xyb[0] || xya[0] > xyb[2]) return(0);
  if (xya[3]<xyb[1] || xya[1] > xyb[3]) return(0);
  return(1);
} /*inside*/

change_src(jd,change,port,new_val)  Jdat *jd;  int change,port;  Rtyp new_val;
/*************************************************************/
/* Change the desired source information.                    */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr30apr91 */
{
  if (port<0 || port>=jd->nport)
    announce_bug("change_src");

  switch (change)
  { case NEW_VM : *(jd->vsrc_mg + port) = new_val;  break;
    case NEW_PH : *(jd->vsrc_ph + port) = new_val;  break;
    case NEW_R  : *(jd->rsrc + port) = new_val;  break;
    case NEW_X  : *(jd->xsrc + port) = new_val;  break;
    case NEW_L  : *(jd->lsrc + port) = new_val;  break;
    case NEW_C  : *(jd->csrc + port) = new_val;  break;
  }
} /*change_src*/

fill_pdat(pd0,td,jd,calc_mask,pvar,pport,pfreq,pn,plist,order)  Pdat *pd0;  Tdat *td;  Jdat *jd;
                       int calc_mask,pvar,pport,pfreq,pn,order;  Rtyp *plist;
/*************************************************************/
/* Fill linked list pd with data, executive routine.         */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr30apr91 */
{ Fdat *fd;
  Pdat *pd;
  int istep,nsize;
  Rtyp old_val,tmp;
  int npts;  /*Number of data points taken for stats*/
  Rtyp *stats;  /*Ptr to list of statitistics*/
  Rtyp rdx,rdy;  /*cell size in meters*/

  /* pd is the plot data structure returned by setup_pdat */
  /* jd is the associated jdat structure */
  /* calc_mask determines what data is calculated:
     CALC_JX|CALC_JY|CALC_EZ for everything*/
  /* pvar is the parameter to be varied:
     STEP_F -- Do all frequencies
     STEP_VM -- Voltage source magnitude
     STEP_VP -- Voltage source phase
     STEP_R -- Source resistance
     STEP_X -- Source reactance
     STEP_L -- Source inductance
     STEP_C -- Source capacitance */
  /* Next four items ignored for STEP_F*/
  /* pport is the port to be varied*/
  /* pfreq is the frequency index*/
  /* pn in the number of points for the pvar parameter*/
  /* plist points to an Rtyp array listing all the parameter values*/
  /* order is for interpolation, ignored for now*/

  /*Update jpr if varying frequency*/
  if (pvar==STEP_F) pn = jd->nfreq;
  else
  { if (pfreq<0 || pfreq>=jd->nfreq) /*Is this a valid freq index?*/
      announce_bug("fill_pdat");
  }

  /*Store old values*/
  switch (pvar)
  { case STEP_VM: old_val = *(jd->vsrc_mg+pport);  break;
    case STEP_VP: old_val = *(jd->vsrc_ph+pport);  break;
    case STEP_R:  old_val = *(jd->rsrc+pport);  break;
    case STEP_X:  old_val = *(jd->xsrc+pport);  break;
    case STEP_L:  old_val = *(jd->lsrc+pport);  break;
    case STEP_C:  old_val = *(jd->csrc+pport);  break;
  }

  /*(Re-)Allocate td data memory*/
  if ( td != NULL && td->tri->n > 0 )
  { nsize = td->tri->n * pn * sizeof(float);
    if ( !(calc_mask&CALC_JX) || pn != td->pn )  /*Free up memory*/
    { if (td->jx_mg != NULL) sFree((char *)td->jx_mg);  td->jx_mg = NULL;
      if (td->jx_ph != NULL) sFree((char *)td->jx_ph);  td->jx_ph = NULL;
    }
    if ( calc_mask&CALC_JX ) /*Allocate memory*/
    { if (td->jx_mg==NULL) td->jx_mg = (float *)new( nsize );
      if (td->jx_ph==NULL) td->jx_ph = (float *)new( nsize );
    }
    if ( !(calc_mask&CALC_JY) || pn != td->pn )  /*Free up memory*/
    { if (td->jy_mg != NULL) sFree((char *)td->jy_mg);  td->jy_mg = NULL;
      if (td->jy_ph != NULL) sFree((char *)td->jy_ph);  td->jy_ph = NULL;
    }
    if ( calc_mask&CALC_JY ) /*Allocate memory*/
    { if (td->jy_mg==NULL) td->jy_mg = (float *)new( nsize );
      if (td->jy_ph==NULL) td->jy_ph = (float *)new( nsize );
    }
    if ( !(calc_mask&CALC_EZ) || pn != td->pn )  /*Free up memory*/
    { if (td->ez_mg != NULL) sFree((char *)td->ez_mg);  td->ez_mg = NULL;
      if (td->ez_ph != NULL) sFree((char *)td->ez_ph);  td->ez_ph = NULL;
    }
    if ( calc_mask&CALC_EZ ) /*Allocate memory*/
    { if (td->ez_mg==NULL) td->ez_mg = (float *)new( nsize );
      if (td->ez_ph==NULL) td->ez_ph = (float *)new( nsize );
    }
    td->pn = pn;
  }
  
  /*(Re-)Allocate pd data memory*/
  pd = pd0;
  while (pd)
  { nsize = pd->nx * pd->ny * pn * sizeof(float);
    if ( !(calc_mask&CALC_JX) || pn != pd->pn )  /*Free up memory*/
    { if (pd->jx_mg != NULL) sFree((char *)pd->jx_mg);  pd->jx_mg = NULL;
      if (pd->jx_ph != NULL) sFree((char *)pd->jx_ph);  pd->jx_ph = NULL;
    }
    if ( calc_mask&CALC_JX ) /*Allocate memory*/
    { if (pd->jx_mg==NULL) pd->jx_mg = (float *)new( nsize );
      if (pd->jx_ph==NULL) pd->jx_ph = (float *)new( nsize );
    }
    if ( !(calc_mask&CALC_JY) || pn != pd->pn )  /*Free up memory*/
    { if (pd->jy_mg != NULL) sFree((char *)pd->jy_mg);  pd->jy_mg = NULL;
      if (pd->jy_ph != NULL) sFree((char *)pd->jy_ph);  pd->jy_ph = NULL;
    }
    if ( calc_mask&CALC_JY ) /*Allocate memory*/
    { if (pd->jy_mg==NULL) pd->jy_mg = (float *)new( nsize );
      if (pd->jy_ph==NULL) pd->jy_ph = (float *)new( nsize );
    }
    if ( !(calc_mask&CALC_EZ) || pn != pd->pn )  /*Free up memory*/
    { if (pd->ez_mg != NULL) sFree((char *)pd->ez_mg);  pd->ez_mg = NULL;
      if (pd->ez_ph != NULL) sFree((char *)pd->ez_ph);  pd->ez_ph = NULL;
    }
    if ( calc_mask&CALC_EZ ) /*Allocate memory*/
    { if (pd->ez_mg==NULL) pd->ez_mg = (float *)new( nsize );
      if (pd->ez_ph==NULL) pd->ez_ph = (float *)new( nsize );
    }
    pd->pn = pn;
    if (pd->stats) sFree((char *)pd->stats);  pd->stats = NULL;
    pd = pd->next;
  }

  /*Allocate storage for max/min numbers*/
  init_pdat_stats(pd0);

  /*Do each instance of parameter to be swept*/
  for (istep=0,stats=pd0->stats; istep<pn; istep++,stats+=NUM_STATS)
  { switch (pvar)
    { case STEP_F:  pfreq = istep;  break;
      case STEP_VM: *(jd->vsrc_mg+pport) = *(plist+istep);  break;
      case STEP_VP: *(jd->vsrc_ph+pport) = *(plist+istep);  break;
      case STEP_R:  *(jd->rsrc+pport) = *(plist+istep);  break;
      case STEP_X:  *(jd->xsrc+pport) = *(plist+istep);  break;
      case STEP_L:  *(jd->lsrc+pport) = *(plist+istep);  break;
      case STEP_C:  *(jd->csrc+pport) = *(plist+istep);  break;
    }

    fd = jd->fdat + pfreq;
    update_jpr(jd,pfreq);

    /*Generate the plot data*/
    pd = pd0;
    npts = 0;
    /*Area, in meters, of 1x1 cell in cell coord.*/
    /*Min cell size is 2xw in cell coord.*/
    rdx = jd->box->A / jd->box->Capm;
    rdy = jd->box->B / jd->box->Capn;
    while(pd)
    { npts += fill_pdat_do(pd,fd,istep,calc_mask,order,stats,rdx,rdy,jd->xterp,jd->yterp);
      pd = pd->next;
    }
    if ( td != NULL && td->tri->n > 0 )
      fill_tdat_do(td,fd,istep,calc_mask,order,rdx,rdy);

    /*Sum x stored in AVE, sum x**2 stored in SIG, convert to AVE, SIG*/
    if (calc_mask&CALC_JX)
    { if (npts<1)
      { *(stats+JX_MG_MAX) = *(stats+JX_MG_MIN) = 0.0;
        *(stats+JX_PH_MAX) = *(stats+JX_PH_MIN) = 0.0;
        *(stats+JX_MG_AVE) = *(stats+JX_PH_AVE) = 0.0;
      }
      else
      { *(stats+JX_MG_AVE) *= 1.0 / npts;
        *(stats+JX_PH_AVE) *= 1.0 / npts;
      }
      if (npts<2) { *(stats+JX_MG_SIG) = *(stats+JX_PH_SIG) = 0.0; }
      else
      { tmp = *(stats+JX_MG_AVE);
        tmp = *(stats+JX_MG_SIG) / (npts - 1) - tmp * tmp;
        *(stats+JX_MG_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
        tmp = *(stats+JX_PH_AVE);
        tmp = *(stats+JX_PH_SIG) / (npts - 1) - tmp * tmp;
        *(stats+JX_PH_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
      }
    }
     
    if (calc_mask&CALC_JY)
    { if (npts<1)
      { *(stats+JY_MG_MAX) = *(stats+JY_MG_MIN) = 0.0;
        *(stats+JY_PH_MAX) = *(stats+JY_PH_MIN) = 0.0;
        *(stats+JY_MG_AVE) = *(stats+JY_PH_AVE) = 0.0;
      }
      else
      { *(stats+JY_MG_AVE) *= 1.0 / npts;
        *(stats+JY_PH_AVE) *= 1.0 / npts;
      }
      if (npts<2) { *(stats+JY_MG_SIG) = *(stats+JY_PH_SIG) = 0.0; }
      else
      { tmp = *(stats+JY_MG_AVE);
        tmp = *(stats+JY_MG_SIG) / (npts - 1) - tmp * tmp;
        *(stats+JY_MG_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
        tmp = *(stats+JY_PH_AVE);
        tmp = *(stats+JY_PH_SIG) / (npts - 1) - tmp * tmp;
        *(stats+JY_PH_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
      }
    }
     
    if (calc_mask&CALC_EZ)
    { if (npts<1)
      { *(stats+EZ_MG_MAX) = *(stats+EZ_MG_MIN) = 0.0;
        *(stats+EZ_PH_MAX) = *(stats+EZ_PH_MIN) = 0.0;
        *(stats+EZ_MG_AVE) = *(stats+EZ_PH_AVE) = 0.0;
      }
      else
      { *(stats+EZ_MG_AVE) *= 1.0 / npts;
        *(stats+EZ_PH_AVE) *= 1.0 / npts;
      }
      if (npts<2) { *(stats+EZ_MG_SIG) = *(stats+EZ_PH_SIG) = 0.0; }
      else
      { tmp = *(stats+EZ_MG_AVE);
        tmp = *(stats+EZ_MG_SIG) / (npts - 1) - tmp * tmp;
        *(stats+EZ_MG_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
        tmp = *(stats+EZ_PH_AVE);
        tmp = *(stats+EZ_PH_SIG) / (npts - 1) - tmp * tmp;
        *(stats+EZ_PH_SIG) = (tmp>TINY) ? sqrt(tmp) : 0.0;
      }
    }

  }

  /*Restore original value of swept parameter*/
  switch (pvar)
  { case STEP_VM: *(jd->vsrc_mg+pport) = old_val;  break;
    case STEP_VP: *(jd->vsrc_ph+pport) = old_val;  break;
    case STEP_R:  *(jd->rsrc+pport) = old_val;  break;
    case STEP_X:  *(jd->xsrc+pport) = old_val;  break;
    case STEP_L:  *(jd->lsrc+pport) = old_val;  break;
    case STEP_C:  *(jd->csrc+pport) = old_val;  break;
  }

} /*fill_pdat*/

fill_pdat_do(pd,fd,istep,mask,order,stats,rdx,rdy,xterp,yterp)  Pdat *pd;  Fdat *fd;
              Rtyp *stats;  Rtyp rdx,rdy;  int istep,mask,order;  int *xterp,*yterp;
/*************************************************************/
/* Fill linked list pd with data, executive routine.         */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr18apr91 */
{ int *ira,*irb;
  int i,j,i_start,i_stop,dat_head;
  Rtyp pta,ptb;
  Rtyp xa,ya,re0,im0,re1,im1,mg0,ph0,mg1,ph1;
  Rtyp a_mg,a_ph,b_mg,b_ph,c_mg,c_ph,d_mg,d_ph;
  Rtyp div_re,div_im;
  float *dmp,*dpp;
  int npts;  /*Number of pts over which statistics are taken*/

  ira = pd->ira;  irb = pd->irb;
  npts = i_start = 0;
  dat_head = pd->nx * pd->ny * istep;
  while (i_start<pd->nx && ira<=irb)
  { xa = pd->x0 + i_start * pd->dx;
    /*Find correct next rectangle (usually present rectangle*/
    while( ira <= irb && *(ira+RX1) < xa+pd->xx0 )
      ira += RSIZE;
    if ( ira > irb ) break;  /*All done*/

    /*figure out how far we can go in this rectangle*/
    i_stop = ( *(ira+RX1) - pd->x0 - pd->xx0 + 1.00001 * pd->dx ) / pd->dx;
    if (i_stop > pd->nx) i_stop = pd->nx;

    /*Calculate the interpolation coeffients*/
    div_re = div_im = 0.0;  /*Accumulate divergence of J*/
    if ( mask&(CALC_JX|CALC_EZ) )
    { if ( *(ira+RXA) >= 0 )
      { re0 = Vec_get_re(*(ira+RXA),fd->jpr)/rdy;
        im0 = Vec_get_im(*(ira+RXA),fd->jpr)/rdy;
        pta = *(xterp + *(ira+RXA));
      }
      else { re0 = im0 = 0.0;  pta = *(ira+RX0); }
      mg0 = hypot(im0,re0);
      ph0 = ( mg0<TINY ) ? 0.0 : atan2(im0,re0);
      if ( *(ira+RXB) >= 0 )
      { re1 = Vec_get_re(*(ira+RXB),fd->jpr)/rdy;
        im1 = Vec_get_im(*(ira+RXB),fd->jpr)/rdy;
        ptb = *(xterp + *(ira+RXB));
      }
      else { re1 = im1 = 0.0;  ptb = *(ira+RX1); }
      mg1 = hypot(im1,re1);
      ph1 = ( mg1<TINY ) ? ph0 : atan2(im1,re1);
      if ( mg0<TINY )  /*No current in 0th rect, so*/
	ph0 = ph1;     /*interpolate to constant phase*/
      if (mg1 > *(stats+JX_MG_MAX)) *(stats+JX_MG_MAX) = mg1;
      if (mg1 < *(stats+JX_MG_MIN)) *(stats+JX_MG_MIN) = mg1;
      if (ph1 > *(stats+JX_PH_MAX)) *(stats+JX_PH_MAX) = ph1;
      if (ph1 < *(stats+JX_PH_MIN)) *(stats+JX_PH_MIN) = ph1;
      *(stats+JX_MG_AVE) += mg1;  *(stats+JX_MG_SIG) += mg1 * mg1;
      *(stats+JX_PH_AVE) += ph1;  *(stats+JX_PH_SIG) += ph1 * ph1;
      a_mg = (mg1-mg0) / ( ptb - pta );
      a_ph = (ph1-ph0) / ( ptb - pta );
      b_mg = ( ptb * mg0 - pta * mg1 ) / ( ptb - pta );
      b_ph = ( ptb * ph0 - pta * ph1 ) / ( ptb - pta );
      div_re = (re1-re0) / (( ptb - pta ) * rdx);
      div_im = (im1-im0) / (( ptb - pta ) * rdx);
    }

    if ( mask&(CALC_JY|CALC_EZ) )
    { if ( *(ira+RYA) >= 0 )
      { re0 = Vec_get_re(*(ira+RYA),fd->jpr)/rdx;
        im0 = Vec_get_im(*(ira+RYA),fd->jpr)/rdx;
        pta = *(yterp + *(ira+RYA));
      }
      else { re0 = im0 = 0.0;  pta = *(ira+RY0); }
      mg0 = hypot(im0,re0);
      ph0 = ( mg0<TINY ) ? 0.0 : atan2(im0,re0);
      if ( *(ira+RYB) >= 0 )
      { re1 = Vec_get_re(*(ira+RYB),fd->jpr)/rdx;
        im1 = Vec_get_im(*(ira+RYB),fd->jpr)/rdx;
        ptb = *(yterp + *(ira+RYB));
      }
      else { re1 = im1 = 0.0;  ptb = *(ira+RY1); }
      mg1 = hypot(im1,re1);
      ph1 = ( mg1<TINY ) ? ph0 : atan2(im1,re1);
      if ( mg0<TINY )  /*No current in 0th rect, so*/
	ph0 = ph1;     /*interpolate to constant phase*/
      if (mg1 > *(stats+JY_MG_MAX)) *(stats+JY_MG_MAX) = mg1;
      if (mg1 < *(stats+JY_MG_MIN)) *(stats+JY_MG_MIN) = mg1;
      if (ph1 > *(stats+JY_PH_MAX)) *(stats+JY_PH_MAX) = ph1;
      if (ph1 < *(stats+JY_PH_MIN)) *(stats+JY_PH_MIN) = ph1;
      *(stats+JY_MG_AVE) += mg1;  *(stats+JY_MG_SIG) += mg1 * mg1;
      *(stats+JY_PH_AVE) += ph1;  *(stats+JY_PH_SIG) += ph1 * ph1;
      c_mg = (mg1-mg0) / ( ptb - pta );
      c_ph = (ph1-ph0) / ( ptb - pta );
      d_mg = ( ptb * mg0 - pta * mg1 ) / ( ptb - pta );
      d_ph = ( ptb * ph0 - pta * ph1 ) / ( ptb - pta );
      div_re += (re1-re0) / (( ptb - pta ) * rdy);
      div_im += (im1-im0) / (( ptb - pta ) * rdy);
    }

    /*Calculate the interpolated data*/
    if ( mask&CALC_JX )
    { dmp = pd->jx_mg + dat_head + i_start; 
      dpp = pd->jx_ph + dat_head + i_start;
      for (i=i_start; i<i_stop; i++,dmp++,dpp++)
      { xa = pd->x0 + pd->xx0 + i * pd->dx;
        *dmp = (float)(a_mg * xa + b_mg);
        *dpp = (float)(a_ph * xa + b_ph);
        for (j=1; j<pd->ny; j++)
        { *(dmp + j * pd->nx) = *dmp;
          *(dpp + j * pd->nx) = *dpp;
        }
      }
    }

    if ( mask&CALC_JY )
    { dmp = pd->jy_mg + dat_head + i_start;
      dpp = pd->jy_ph + dat_head + i_start;
      for (j=0; j<pd->ny; j++,dmp+=pd->nx,dpp+=pd->nx)
      { ya = pd->y0 + pd->yy0 + j * pd->dy;
        *dmp = (float)(c_mg * ya + d_mg);
        *dpp = (float)(c_ph * ya + d_ph);
        for (i=1; i<i_stop-i_start; i++)
        { *(dmp + i) = *dmp;
          *(dpp + i) = *dpp;
        }
      }
    }

    if ( mask&CALC_EZ )
    { dmp = pd->ez_mg + dat_head;
      dpp = pd->ez_ph + dat_head;
      /*using b as tmp for calculating divergence*/
      /* Q = j * div(J) / W ) */
      b_mg = 1.0 / ( 1.0e6 * TWO_PI * fd->freq );
      re0 = - div_im * b_mg;
      im0 = div_re * b_mg;
      b_mg = hypot(im0,re0);
      b_ph = (b_mg>TINY) ? atan2(im0,re0) : 0.0;
      if (b_mg > *(stats+EZ_MG_MAX)) *(stats+EZ_MG_MAX) = b_mg;
      if (b_mg < *(stats+EZ_MG_MIN)) *(stats+EZ_MG_MIN) = b_mg;
      if (b_ph > *(stats+EZ_PH_MAX)) *(stats+EZ_PH_MAX) = b_ph;
      if (b_ph < *(stats+EZ_PH_MIN)) *(stats+EZ_PH_MIN) = b_ph;
      *(stats+EZ_MG_AVE) += b_mg;  *(stats+EZ_MG_SIG) += b_mg * b_mg;
      *(stats+EZ_PH_AVE) += b_ph;  *(stats+EZ_PH_SIG) += b_ph * b_ph;
      for (i=i_start; i<i_stop; i++)
        for (j=0; j<pd->ny; j++)
        { *(dmp + i + j * pd->nx) = (float)b_mg;
          *(dpp + i + j * pd->nx) = (float)b_ph;
        }
    }

    npts += 1;
    /*inc to the next rectangle*/
    i = i_start = i_stop;
    ira += RSIZE;
  }
  return(npts);
} /*fill_pdat_do*/

fill_tdat_do(td,fd,istep,mask,order,rdx,rdy)  Tdat *td;  Fdat *fd;
			  int istep,mask,order;  Rtyp rdx,rdy;
/*************************************************************/
/* Fill td with current data.                                */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01jun93 */
{ Ipts *tri;
  int *ip;
  int i,dat_head,loc;
  Rtyp re0,im0,ph0,mg0;

  if ( td == NULL )
    return;

  tri = td->tri;
  dat_head = td->tri->n * istep;  /*istep-th freq*/

  for (i=0; i<tri->n; i++)
  { ip = tri->ipt0 + tri->size * i;
    loc = *( ip + TA0 );

    re0 = Vec_get_re(loc,fd->jpr);
    im0 = Vec_get_im(loc,fd->jpr);
    mg0 = hypot(im0,re0);
    ph0 = ( mg0<TINY ) ? 0.0 : atan2(im0,re0);
    if ( mask & CALC_JX )
    { *(td->jx_mg + dat_head + i) = mg0/rdy;
      *(td->jx_ph + dat_head + i) = ph0;
    }
    if ( mask & CALC_JY )
    { *(td->jy_mg + dat_head + i) = mg0/rdx;
      *(td->jy_ph + dat_head + i) = ph0;
    }
    if ( mask & CALC_EZ )
    { *(td->ez_mg + dat_head + i) = 0.0;
      *(td->ez_ph + dat_head + i) = 0.0;
    }
  }
} /*fill_tdat_do*/

Tdat *setup_tdat(jd,ilev,xy,dx,dy)  Jdat *jd;  int ilev;  Rtyp xy[4],dx,dy;
/*************************************************************/
/* Return Tdat structure ready for a to be calc'ed.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr18apr91 */
{ Tdat *td;
  Rtyp rtt,x0,y0;
  int ikey[TSIZE];
  msub_list *lev;
  msub *im,*mstop;
  int itt;

  /*ilev = level number,
    xy[0] = x0, xy[1] = y0, xy[2] = x1, xy[3] = y1.
    if x0>x1 or y0>y1, order will be swapped.
    coordinates are 2 unit per cell.
    xy is region for returning triangle data in td.
  */

  /*Can't specify ground level (cir-n-1) for plotting*/
  if (ilev<0 || ilev>=jd->cir->n-1)
    announce_bug("setup_tdat(0)");

  if (xy[0]==xy[2] || xy[1]==xy[3]) return(NULL);

  /*Ck to make sure region pts are in rite order*/
  if (xy[0]>xy[2]) { rtt = xy[0];  xy[0] = xy[1];  xy[1] = rtt; }
  if (xy[1]>xy[3]) { rtt = xy[1];  xy[1] = xy[3];  xy[3] = rtt; }

  td = (Tdat *)cnew( 1,sizeof(Tdat) );
  init_tdat(td);

  td->tri = (Ipts *)cnew(1,sizeof(Ipts));
  init_ipts(td->tri,TSIZE);

  itt = (xy[0] + 0.99999*dx) / dx;
  td->x0 = itt * dx;
  if (td->x0 < 0.0) td->x0 = 0.0;

  itt = (xy[1] + 0.99999*dy) / dy;
  td->y0 = itt * dy;
  if (td->y0 < 0.0) td ->y0 = 0.0;

  td->dx = dx;
  td->dy = dy;

  lev = jd->cir->lev0 + ilev;

  /*Get info on each triangle within desired region*/
  mstop = lev->msub0 + lev->n;
  for (im=lev->msub0; im<mstop; im++)
  { x0 = (Rtyp)(*im->ip);
    y0 = (Rtyp)(*(im->ip+1));
    if ( ( im->id == MTNN || im->id == MTNP || im->id == MTPN || im->id == MTPP ) &&
	   x0 > xy[0]-1.0 && x0 < xy[2]+1.0 && y0 > xy[1]-1.0 && y0 < xy[3]+1.0 )
    { switch (im->id)
      { case MTPP : ikey[TX0] = x0 - 1;  ikey[TY0] = y0 - 1;        /*low x diagonal point*/
		    ikey[TX1] = x0 + 1;  ikey[TY1] = y0 + 1;        /*high x diagonal point*/
		    ikey[TX2] = ikey[TX0];  ikey[TY2] = ikey[TY1];  /*Rt. angle vertex point*/
		    break;
        case MTPN : ikey[TX0] = x0 - 1;  ikey[TY0] = y0 + 1;        /*low x diagonal point*/
		    ikey[TX1] = x0 + 1;  ikey[TY1] = y0 - 1;        /*high x diagonal point*/
		    ikey[TX2] = ikey[TX0];  ikey[TY2] = ikey[TY1];  /*Rt. angle vertex point*/
		    break;
        case MTNN : ikey[TX0] = x0 - 1;  ikey[TY0] = y0 - 1;        /*low x diagonal point*/
		    ikey[TX1] = x0 + 1;  ikey[TY1] = y0 + 1;        /*high x diagonal point*/
		    ikey[TX2] = ikey[TX1];  ikey[TY2] = ikey[TY0];  /*Rt. angle vertex point*/
		    break;
        case MTNP : ikey[TX0] = x0 - 1;  ikey[TY0] = y0 + 1;        /*low x diagonal point*/
		    ikey[TX1] = x0 + 1;  ikey[TY1] = y0 - 1;        /*high x diagonal point*/
		    ikey[TX2] = ikey[TX1];  ikey[TY2] = ikey[TY0];  /*Rt. angle vertex point*/
		    break;
      }
      ikey[TA0] = im->loc;                      /*location of current data*/
      add_ipts(td->tri,ikey);
    }
  }

  return(td);

} /*setup_tdat*/

snap_tdat(pd0,td)  Pdat *pd0;  Tdat *td;
/*************************************************************/
/* Snap the triangle coord in td to corners of rects in pd.  */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr08jun93 */
{ Pdat *pd;
  Rtyp pc[RSIZE],tc[TSIZE];  /*rect and tri coordinates*/
  int i,snap_cnt,dist0,dist1;
  int *ip;

  /*This proceedure assumes that pd and jd were set up with same*/
  /*parameters, especially, dx,dy and xy[*].*/

  if ( pd0 == NULL || td == NULL )  /*No rect or triangles*/
    return;

  if ( td->tri->n <= 0 )  /*No triangles to snap*/
    return;

  if ( td->snapped ) /*Triangle coord already converted to pixel coord*/
    return;

  td->snapped = 1;

  /*Convert all triangle vertices to pixel coordinates rel. to pd->xx0,yy0*/
  for (i=0; i<td->tri->n; i++)
  { ip = td->tri->ipt0 + i * td->tri->size;
    /*Do first vertice*/
    *(ip+TX0) = 0.5 + (*(ip+TX0) - pd0->xx0) / pd0->dx;
    *(ip+TY0) = 0.5 + (*(ip+TY0) - pd0->yy0) / pd0->dy;
    /*Do second vertice*/
    *(ip+TX1) = 0.5 + (*(ip+TX1) - pd0->xx0) / pd0->dx;
    *(ip+TY1) = 0.5 + (*(ip+TY1) - pd0->yy0) / pd0->dy;
    /*Do third vertice*/
    *(ip+TX2) = 0.5 + (*(ip+TX2) - pd0->xx0) / pd0->dx;
    *(ip+TY2) = 0.5 + (*(ip+TY2) - pd0->yy0) / pd0->dy;
    /*Make absolutely sure the third vertex makes a rt. angle triangle*/
    if ( *(ip+TX2) != *(ip+TX0) && *(ip+TX2) != *(ip+TX1) )
    { dist0 = *(ip+TX0) - *(ip+TX2);
      if ( dist0 < 0 ) dist0 = -dist0;
      dist1 = *(ip+TX1) - *(ip+TX2);
      if ( dist1 < 0 ) dist1 = -dist1;
      *(ip+TX2) = ( dist0 < dist1 ) ? *(ip+TX0) : *(ip+TX1);
    }
    if ( *(ip+TY2) != *(ip+TY0) && *(ip+TY2) != *(ip+TY1) )
    { dist0 = *(ip+TY0) - *(ip+TY2);
      if ( dist0 < 0 ) dist0 = -dist0;
      dist1 = *(ip+TY1) - *(ip+TY2);
      if ( dist1 < 0 ) dist1 = -dist1;
      *(ip+TY2) = ( dist0 < dist1 ) ? *(ip+TY0) : *(ip+TY1);
    }
  }

  /*Do not snap if pixel is bigger than 0.4 cell*/
  if ( pd0->dx > 0.8 || pd0->dy > 0.8 )
    return;

  /*Go thru each pd rectangle*/
  pd = pd0;
  while ( pd != NULL )
  { /*Calculate pixel coord of rect. corners relative to pd->xx0,yy0*/
    i = (int)(pd->x0/pd->dx+0.5);    /*Low coord pt*/
    pc[RX0] = (Rtyp)i;
    i = (int)(pd->y0/pd->dy+0.5);
    pc[RY0] = (Rtyp)i;
    pc[RX1] = pc[RX0] + (pd->nx-1); /*High coord pt*/
    pc[RY1] = pc[RY0] + (pd->ny-1);
    
    /*Now search thru the triangles for any vertices within one pixel*/
    /*of the rectangle corners.  If found, snap them so they are next*/
    /*to the rectangle, no overlap, no blank space.*/

    snap_cnt = 0;
    for (i=0; i<td->tri->n && snap_cnt<2; i++)
    { ip = td->tri->ipt0 + i * td->tri->size;
      /*get the diagonal vertice values and convert to float*/
      tc[TX0] = *(ip+TX0);  tc[TY0] = *(ip+TY0);  /*right most diag pt*/
      tc[TX1] = *(ip+TX1);  tc[TY1] = *(ip+TY1);  /*left most diag pt*/
      tc[TX2] = *(ip+TX2);  tc[TY2] = *(ip+TY2);  /*rt angle pt*/
      if ( do_tdat_snap(pc,tc) )
      { *(ip+TX0) = tc[TX0];  *(ip+TY0) = tc[TY0];
        *(ip+TX1) = tc[TX1];  *(ip+TY1) = tc[TY1];
        *(ip+TX2) = tc[TX2];  *(ip+TY2) = tc[TY2];
	snap_cnt++;
      }
    }
    pd = pd->next;
  }
} /*snap_tdat*/

#define SNAP_LIMIT 2.1

do_tdat_snap(pc,tc)  Rtyp pc[RSIZE],tc[TSIZE];
/*************************************************************/
/* Do the actual snapping for a given rect and tri coord.    */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr08jun93 */
{
  /*Check the right most diagonal vertice against first rect corner*/
  if ( fabs(tc[TX0]-pc[RX0]) <= SNAP_LIMIT &&
       fabs(tc[TY0]-pc[RY0]) <= SNAP_LIMIT    )
  { tc[TX0] = pc[RX0];
    tc[TY0] = tc[TY2] = pc[RY0] - 1.0;
  }

  /*Check the right most diagonal vertice against second rect corner*/
  else if ( fabs(tc[TX0]-pc[RX0]) <= SNAP_LIMIT &&
	    fabs(tc[TY0]-pc[RY1]) <= SNAP_LIMIT    )
  { tc[TX0] = pc[RX0];
    tc[TY0] = tc[TY2] = pc[RY1] + 1.0;
  }

  /*Check the right most diagonal vertice against third rect corner*/
  else if ( fabs(tc[TX0]-pc[RX1]) <= SNAP_LIMIT && 
	    fabs(tc[TY0]-pc[RY1]) <= SNAP_LIMIT &&
	    tc[TY1] < tc[TY0]                      )
  { tc[TX0] = tc[TX2] = pc[RX1] + 1.0;
    tc[TY0] = pc[RY1];
  }

  /*Check the right most diagonal vertice against fourth rect corner*/
  else if ( fabs(tc[TX0]-pc[RX1]) <= SNAP_LIMIT &&
	    fabs(tc[TY0]-pc[RY0]) <= SNAP_LIMIT &&
            tc[TY1] > tc[TY0]                      )
  { tc[TX0] = tc[TX2] = pc[RX1] + 1.0;
    tc[TY0] = pc[RY0];
  }

  /*Check the left most diagonal vertice against fourth rect corner*/
  else if ( fabs(tc[TX1]-pc[RX1]) <= SNAP_LIMIT &&
	    fabs(tc[TY1]-pc[RY0]) <= SNAP_LIMIT    )
  { tc[TX1] = pc[RX1];
    tc[TY1] = tc[TY2] = pc[RY0] - 1.0;
  }

  /*Check the left most diagonal vertice against third rect corner*/
  else if ( fabs(tc[TX1]-pc[RX1]) <= SNAP_LIMIT &&
	    fabs(tc[TY1]-pc[RY1]) <= SNAP_LIMIT    )
  { tc[TX1] = pc[RX1];
    tc[TY1] = tc[TY2] = pc[RY1] + 1.0;
  }

  /*Check the left most diagonal vertice against second rect corner*/
  else if ( fabs(tc[TX1]-pc[RX0]) <= SNAP_LIMIT &&
	    fabs(tc[TY1]-pc[RY1]) <= SNAP_LIMIT &&
            tc[TY1] > tc[TY0]                      )
  { tc[TX1] = tc[TX2] = pc[RX0] - 1.0;
    tc[TY1] = pc[RY1];
  }

  /*Check the left most diagonal vertice against first rect corner*/
  else if ( fabs(tc[TX1]-pc[RX0]) <= SNAP_LIMIT &&
	    fabs(tc[TY1]-pc[RY0]) <= SNAP_LIMIT &&
            tc[TY1] < tc[TY0]                      )
  { tc[TX1] = tc[TX2] = pc[RX0] - 1.0;
    tc[TY1] = pc[RY0];
  }

  else
    return(0);

  return(1);
} /*do_tdat_snap*/

Jdat *read_jdat(file_name)  char *file_name;
/*************************************************************/
/* Oversee reading of jxy file.                              */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ FILE *in;
  Jdat *jd;
  int i;
  Jdat *get_jd();

  if ( file_name==NULL )
    announce_bug("read_jdat");

  if ( (in=fopen(file_name,"r")) == NULL )
  { fprintf(stderr,"\n%s: cannot read file %s.\n",prog_name,file_name);
    return(NULL);
  }

  Fgets_reset();

  find_next(in,"JHD");
  /*Read each section of jxy (one per circuit analysis)*/
  jd = get_jd(file_name,in);

  fclose(in);

  if ( jd == NULL ) return (NULL);		/* KRS */
  setup_jdat(jd);  /*finish intializing it*/

  return(jd);
} /*read_jdat*/

Jdat *get_jd(file_name,in)  char *file_name;  FILE *in;
/*************************************************************/
/* Read in next section of jxy file, recurse for more sec's. */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ int i,j;
  char str[256];
  char *pos;
  int nsub,ilev,loc,id,ip[6];
  msub_list *lev;
  Fdat *fd,*fd0;
  Jdat *jd;
  Fdat *get_fd();

  /*As passed to this procedure, a "JHD" line must have just
    been read*/

  if (sFgets(str,255,in)==NULL) return(NULL);  /*EOF*/
  pos = Skip_blnks(str);

  /*Open up room for jd and initialize it*/
  jd = (Jdat *)new(sizeof(Jdat));
  init_jdat(jd);
  
  /*Keep going until GEO line seen*/
  while (strncmp(pos,"GEO",3)!=0)
  { if ( strncmp(pos,"CMD",3)==0  && jd->cmd == NULL )
    { pos = Next_field(pos);
      jd->cmd = cnew(strlen(pos)+1,sizeof(char));
      strncpy(jd->cmd,pos,strlen(pos));
    }
    if ( strncmp(pos,"TIM",3)==0 && jd->tim == NULL )
    { pos = Next_field(pos);
      jd->tim = cnew(strlen(pos)+1,sizeof(char));
      strncpy(jd->tim,pos,strlen(pos));
    }
    if ( strncmp(pos,"VER",3)==0 && jd->ver == NULL )
    { pos = Next_field(pos);
      jd->ver = cnew(strlen(pos)+1,sizeof(char));
      strncpy(jd->ver,pos,strlen(pos));
    }
    if ( strncmp(pos,"NAM",3)==0 && jd->nam == NULL )
    { pos = Next_field(pos);
      jd->nam = cnew(strlen(pos)+1,sizeof(char));
      strncpy(jd->nam,pos,strlen(pos));
    }
    if ( strncmp(pos,"JHD",3)==0 ) /*clear this jd and start new one*/
    { clear_jdat(jd);
      sFree((char *)jd);
      return(get_jd(file_name,in));
    }
    if ( sFgets(str,255,in) == NULL )
    { clear_jdat(jd);
      sFree((char *)jd);
      return(NULL); /*EOF*/
    }
    pos = Skip_blnks(str);
  }

  /*We are at the .geo file image, read it in*/
  if ( read_cir(file_name,in,jd->box,jd->cir,0) < 0 )
  { clear_jdat(jd);
    sFree((char *)jd);
    find_next(in,"JHD");
    return(get_jd(file_name,in));
  }

  xfrm_ports(jd->cir->ports);

  jd->nport = jd->cir->ports->m;
  jd->vsrc_mg = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );
  jd->vsrc_ph = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );
  jd->rsrc = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );
  jd->xsrc = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );
  jd->lsrc = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );
  jd->csrc = (Rtyp *)cnew( jd->nport,sizeof(Rtyp) );

  /*Finished with .geo, find a SUB line and allocate memory*/
  while ( strncmp(pos,"SUB",3)!=0 )
  { if ( strncmp(pos,"JHD",3)==0 ) /*clear this jd and start new one*/
    { clear_jdat(jd);
      sFree((char *)jd);
      return(get_jd(file_name,in));
    }
    if( sFgets(str,255,in) == NULL )
    { clear_jdat(jd);
      sFree((char *)jd);
      return(NULL); /*EOF*/
    }
    pos = Skip_blnks(str);
  }

  /*Read the subsection list*/
  jd->nsub = 0;
  while ( strncmp(pos,"SUB",3)==0 )
  { /*Get the level number and number of subsections*/
    ilev = atoi(pos=Next_field(pos));
    nsub = atoi(Next_field(pos));
    if (ilev<0 || ilev>=jd->cir->n || nsub<0 )
    { fprintf(stderr,"\n%s: File %s, SUB %d %d bad, ignoring %s.\n",
                  prog_name,jd->box->file,ilev,nsub,jd->nam);
      clear_jdat(jd);
      sFree((char *)jd);
      find_next(in,"JHD");
      return(get_jd(file_name,in));
    }

    lev = jd->cir->lev0 + ilev;
    jd->nsub += nsub;

    for (i=0; i<nsub; i++)
    { sFgets(str,255,in);
      pos = Skip_blnks(str);
      if ( strncmp(pos,"JHD",3)==0 )
      { fprintf(stderr,"\n%s: File %s, SUB %d %d, wrong number of subs, ignoring %s.\n",
                    prog_name,jd->box->file,ilev,nsub,jd->nam);
        clear_jdat(jd);
        sFree((char *)jd);
        return(get_jd(file_name,in));
      }
      loc = atoi( pos );
      id  = atoi( pos = Next_field(pos) );
      ip[0] = atoi( pos = Next_field(pos) );
      ip[1] = atoi( pos = Next_field(pos) );
      if (id >= NUM_GSUB)
      { ip[2] = atoi( pos = Next_field(pos) );
        ip[3] = atoi( pos = Next_field(pos) );
        ip[4] = atoi( pos = Next_field(pos) );
        ip[5] = atoi( Next_field(pos) );
      }
      add_msub(id,ip,jd->box,lev);
      (lev->msub0+lev->n-1)->loc = loc;
    }

    /*Next line must be an END*/
    sFgets(str,255,in);
    pos = Skip_blnks(str);
    if ( strncmp(pos,"END",3)!=0 )
    { fprintf(stderr,"\n%s: File %s, SUB %d %d, wrong number of subs, ignoring %s.\n",
                  prog_name,jd->box->file,ilev,nsub,jd->nam);
      clear_jdat(jd);
      sFree((char *)jd);
      if (strncmp(pos,"JHD",3)!=0) find_next(in,"JHD");
      return(get_jd(file_name,in));
    }

    sFgets(str,255,in);
    pos = Skip_blnks(str);
  }
  if ( strncmp(pos,"END",3)!=0 )
  { fprintf(stderr,"\n%s: File %s, SUB section END bad, ignoring %s.\n",
                  prog_name,jd->box->file,jd->nam);
    clear_jdat(jd);
    sFree((char *)jd);
    if (strncmp(pos,"JHD",3)!=0) find_next(in,"JHD");
    return(get_jd(file_name,in));
  }

  /*Allocate storage for interpolation coordinates*/
  jd->xterp = (int *)cnew(jd->nsub,sizeof(int));
  jd->yterp = (int *)cnew(jd->nsub,sizeof(int));

  /*Read in frequency data*/
  fd0 = get_fd(in,jd);

  /*Count the number of frequency data sections and set up freq list*/
  jd->nfreq = 0;
  fd = fd0;
  while (fd) { jd->nfreq++;  fd = fd->next; }
  if (fd0) /*Set up freq list only if there is freq data*/
  { jd->freq = (Rtyp *)new( jd->nfreq*sizeof(Rtyp) );
    jd->fdat = (Fdat *)new( jd->nfreq*sizeof(Fdat) );

    /*Organize fd data into array*/
    for (i=0,fd=fd0; i<jd->nfreq; i++,fd=fd->next)
    { copy_fdat(jd->fdat+i,fd);
      (jd->fdat+i)->next = jd->fdat+i+1;
      *(jd->freq+i) = fd->freq;
    }
    (jd->fdat+jd->nfreq-1)->next = NULL;
    while (fd0) { fd = fd0;  fd0 = fd0->next;  sFree((char *)fd); }
  }

  /*Now try to get another section*/
  jd->next = get_jd(file_name,in);
  return(jd);
} /*get_jd*/

Fdat *get_fd(in,jd)  FILE *in;  Jdat *jd;
/*************************************************************/
/* Read in next next freq. section, recurse for more sec's.  */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ int i,j;
  char str[256];
  char *pos;
  Fdat *fd;
  Rtyp re,im;
  Rtyp rdx,rdy;

  rdx = 0.5 * jd->box->A / jd->box->Capm;
  rdy = 0.5 * jd->box->B / jd->box->Capn;

  /*File must be positioned BEFORE FRE line*/

  if (sFgets(str,255,in)==NULL) return(NULL);  /*EOF*/
  pos = Skip_blnks(str);

  /*Keep going until FRE line seen or JHD line indicates last freq*/
  while (strncmp(pos,"FRE",3)!=0)
  { if ( strncmp(pos,"JHD",3)==0 ) return(NULL);
    if ( sFgets(str,255,in)==NULL ) return(NULL);
    pos = Skip_blnks(str);
  }

  /*Open up room for fd and initialize it*/
  fd = (Fdat *)new(sizeof(Fdat));
  init_fdat(fd,jd->nport,jd->nsub);

  /*Read the frequency*/
  fd->freq = atof( Next_field(pos) );

  if (sFgets(str,255,in)==NULL) return(NULL);  /*EOF*/
  pos = Skip_blnks(str);

  /*Keep going until an SPR line comes up*/
  while (strncmp(pos,"SPR",3)!=0)
  { if (strncmp(pos,"JHD",3)==0)
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    if (sFgets(str,255,in)==NULL)  /*EOF*/
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    pos = Skip_blnks(str);
  }

  for (i=0; i<jd->nport; i++) for (j=0; j<jd->nport; j++)
  { sFgets(str,255,in);
    pos = Skip_blnks(str);
    if (strncmp(pos,"JHD",3)==0)
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    re = atof(pos);
    im = atof(Next_field(pos));
    Mat_put(i,j,re,im,fd->cans);
  }
  sFgets(str,255,in);
  pos = Skip_blnks(str);
  if ( strncmp(pos,"END",3)!=0)
  { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
  stoy(fd->cans);  /*Leave it in Y-parameters*/

  /*Keep going until a JXY line comes up*/
  while (strncmp(pos,"JXY",3)!=0)
  { if (strncmp(pos,"JHD",3)==0)
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    if (sFgets(str,255,in)==NULL)  /*EOF*/
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    pos = Skip_blnks(str);
  }

  for (i=0; i<jd->nport; i++) for (j=0; j<jd->nsub; j++)
  { sFgets(str,255,in);
    pos = Skip_blnks(str);
    if (strncmp(pos,"JHD",3)==0)
    { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
    re = atof(pos);
    im = atof(Next_field(pos));
    Mat_put(i,j,re,im,fd->ypr);
  }

  if (sFgets(str,255,in)==NULL) /*EOF*/
  { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }
  pos = Skip_blnks(str);
  if ( strncmp(pos,"END",3)!=0 )
  { clear_fdat(fd);  sFree((char *)fd);  return(NULL); }

  /*All current data is current density, except triangles*/
  /*Convert all data to current, not current density*/
  den2current(fd->ypr,jd);

  fd->next = get_fd(in,jd);
  return(fd);
} /*get_fd*/

den2current(ypr,jd)  Mat_ptr ypr;  Jdat *jd;
/*************************************************************/
/* Convert all ypr data from current density to current.     */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr07jun93 */ 
{ int i,ilev,loc;
  int *jr;
  Rtyp dx,dy,conv;
  msub_list *lev;
  msub *im,*mstop;

  /*Size of one cell in meters*/
  dx = jd->box->A / jd->box->Capm;
  dy = jd->box->B / jd->box->Capn;

  /*Go thru each msub on each level and normalize ypr to dx or dy*/
  /*This makes the resulting value as the current density in Amps/cell_coord*/
  for (ilev=0; ilev<jd->cir->n; ilev++)
  { lev = jd->cir->lev0 + ilev;
    mstop = lev->msub0 + lev->n;
    for (im=lev->msub0; im<mstop; im++)
    { jr = im->ip;
      loc = im->loc;
      switch(im->id)
      { case MXR :
	  conv = dy;
	  /* conv *= ( *(jr+MY1) - *(jr+MY0) ) * 0.5; */
          break;
        case MYR :
	  conv = dx;
	  /* conv *= ( *(jr+MX1) - *(jr+MX0) ) * 0.5; */
          break;
        case MTNN :
        case MTNP :
        case MTPN :
        case MTPP :
	  conv = 0.5;
	  break;
        case MENDYN :
        case MENDYP :
	  conv = dx;
          break;
        case MENDXN :
        case MENDXP :
	  conv = dy;
          break;
#ifdef ALPHA
	/* Added to alpha for 2.4b */
	default :
          conv = 0.0;
          break;
#endif /* ALPHA */
        } /*swich im->id*/
	for (i=0; i<ypr->Num_row; i++)  /*index thru each port*/
	  Mat_put(i,loc,conv*Mat_get_re(i,loc,ypr),conv*Mat_get_im(i,loc,ypr),ypr);
      } /*for im*/
  } /*for ilev*/
} /*den2current*/

find_next(in,str2)  FILE *in;  char *str2;
/*************************************************************/
/* Position the file, in, to the next occurance of str2.     */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11apr91 */ 
{ char str[256];

  while ( sFgets(str,255,in) != NULL )
    if ( strncmp(Skip_blnks(str),str2,strlen(str2)) == 0 ) return;

} /*find_next*/

/*Dummy routines for reading .geo file*/

clean_sub(sub)  sub_typ *sub;
{ int i; i=0; }

init_sub_typ(Box)  Box_info *Box;
{ int i; i=0; }

init_ej(Box)  Box_info *Box;
{ int i; i=0; }
